/**********************************************************************************************/
/* FBL_BitSetIterator.cpp																	  */
/*                                                                       					  */
/* Foundation Base Library ( FBL )		                                   					  */
/*                                                                       					  */
/* Copyright Paradigma, 1993-2002                                        					  */
/* All Rights Reserved.                                                 					  */
/**********************************************************************************************/

#ifndef __FBL_BitIterator_h
   #define __FBL_BitIterator_h

#if VSDK_FRAMEWORK_STYLE
#	include <VSDK/FBL_SetIterator.h>
#	include <VSDK/FBL_BitSet.h>
#else
#	include <FBL_SetIterator.h>
#	include <FBL_BitSet.h>
#endif // VSDK_FRAMEWORK_STYLE

#if defined(__MWERKS__)
	#pragma options align=power
	#pragma inline_depth(8)
#endif


/**********************************************************************************************/
#define BITS_IN_CHAR	8


/**********************************************************************************************/
class FBL_BitSetIterator : public FBL_SetIterator
{
	public://///////////////////////////////////////////////////////////////////////////////////

FBL_EXP							FBL_BitSetIterator( const FBL_BitSet& inBitSet );


		/* Iteration methods */

FBL_EXP virtual ID_TYPE	 		operator* ( void ) const 	{ return mCurrentPosition; }
FBL_EXP inline ID_TYPE	 		GetPosition( void )			{ return mCurrentPosition; }

FBL_EXP virtual bool			First( void )	{ return FirstBit_1(); 	}
FBL_EXP virtual bool			Last( void )	{ return LastBit_1();	}

FBL_EXP virtual bool			Next( void )	{ return NextBit_1(); 	}
FBL_EXP virtual bool			Prev( void )	{ return PrevBit_1(); 	}

								// Locate Nth 1-bit. Not very fast function.
FBL_EXP	virtual bool			GoTo( ulong inIndex );


		/* High level methods */

virtual ID_TYPE					FirstItem( void );
virtual ID_TYPE					LastItem( void );
							
virtual ID_TYPE					NextItem( void );
virtual ID_TYPE					PrevItem( void );


	public://///////////////////////////////////////////////////////////////////////////////////
	
		/* BitSet specific methods: */
								// Set iterator to the first 1/0 bit.
								// Returns true if such bit found.
FBL_EXP	bool					FirstBit_1( void );
FBL_EXP	bool					FirstBit_0( void );

								// Move iterator to the last 1/0 bit.
								// Returns true if such bit found.
FBL_EXP	bool					LastBit_1( void );
FBL_EXP	bool					LastBit_0( void );

								// Move iterator to the next 1/0 bit.
								// Returns true if such bit found.
FBL_EXP	bool					NextBit_1( void );
FBL_EXP	bool					NextBit_0( void );

								// Move iterator to the next 1/0 bit.
								// Returns true if such bit found.
FBL_EXP	bool					PrevBit_1( void );
FBL_EXP	bool					PrevBit_0( void );


	public://///////////////////////////////////////////////////////////////////////////////////
	
								// Move iterator to the first/last bit.
FBL_EXP void					FirstBit( void );
FBL_EXP void					LastBit( void );

								// Move iterator to the next/prev bit.
								// Returns true if such bit found.
FBL_EXP	bool					NextBit( void );
FBL_EXP	bool					PrevBit( void );

								// Set iterator to the Nth bit of iterated BitSet.
FBL_EXP bool					GoToBit( ulong inNth );
	
	
    protected://////////////////////////////////////////////////////////////////////////////////

const	FBL_BitSet&				mBitSet;
		uchar*					pEndByte;

		uchar*					pCurrentByte;
		uchar					currBit;			// range [0..31]
		ID_TYPE					mCurrentPosition;
};


/**********************************************************************************************/
inline void FBL_BitSetIterator::FirstBit( void )
{
	pCurrentByte 		= mBitSet.mArray;
	currBit 			= 0;
	mCurrentPosition   	= 1;
}


/**********************************************************************************************/
inline void FBL_BitSetIterator::LastBit( void )
{
	ulong byte = (mBitSet.mMaxValue-1) / 8;

	pCurrentByte 		= mBitSet.mArray + byte;
	currBit 			= (uchar)((mBitSet.mMaxValue-1) % 8);
	mCurrentPosition   	= mBitSet.mMaxValue;
}


/**********************************************************************************************/
inline bool FBL_BitSetIterator::GoTo( ulong inIndex )
{
	CHECK_FBL( inIndex > 0 ); 	

	mCurrentPosition = mBitSet.FindNth_1_Bit( inIndex );
	return mCurrentPosition ? true : false;
}




#pragma mark -
 

/***********************************************************************************************
FUNCTION   : bool FBL_BitSetIterator::NextBit( void );

CALLED BY  : FBL_BitSetIterator::NextBit_1();
			 FBL_BitSetIterator::NextBit_0();	

DESCRIPTION: Move iterator to the next bit. Returns true if such bit found.
***********************************************************************************************/
inline bool FBL_BitSetIterator::NextBit( void )
{
	CHECK_FBL( 1 <= mCurrentPosition && mCurrentPosition <= mBitSet.mMaxValue ); 

	// Work out whether we are at the real end of our array
	if( mCurrentPosition >= mBitSet.mMaxValue )
		return false;
	
	// We are going to move on one index whatever
	mCurrentPosition++;

	// at first try find next bit in current byte
	if( currBit == BITS_IN_CHAR - 1 )
	{
		// now try to find next byte
		if( pCurrentByte < pEndByte )
		{
			pCurrentByte++;
			currBit = 0;			
			return true;
		}
		else
		{
			return false;
		}
	}
	else
	{
		currBit++;
		return true;
	}
}


/***********************************************************************************************
FUNCTION   : bool FBL_BitSetIterator::PrevBit( void );

CALLED BY  : FBL_BitSetIterator::PrevBit_1();
			 FBL_BitSetIterator::PrevBit_0();	

DESCRIPTION: Move iterator to the prev bit. Returns true if such bit found.
***********************************************************************************************/
inline bool FBL_BitSetIterator::PrevBit( void )
{
	CHECK_FBL( 1 <= mCurrentPosition && mCurrentPosition <= mBitSet.mMaxValue ); 

	if( mCurrentPosition <= 1 )
		return false;
	
	// We are going to move on one index whatever
	mCurrentPosition--;

	// at first try find next bit in current byte
	if( currBit == 0 )
	{
		// now try to find next byte
		if( mBitSet.mArray < pCurrentByte )
		{
			pCurrentByte--;
			currBit = BITS_IN_CHAR - 1;			
			return true;
		}
		else
		{
			return false;
		}
	}
	else
	{
		currBit--;
		return true;
	}
}


#pragma mark -
 

/**********************************************************************************************/
inline bool FBL_BitSetIterator::FirstBit_1( void )
{
	if( mBitSet.mBytesInArray == 0 ) // No bits in the array
		return false;
	
	// Move to the beginning of the array
	FirstBit();
	
	// Return the next set bit
	if( (*pCurrentByte) & 0x01 )
	{
		// First bit is set
		return true;
	}
	else
	{
		// Return the next bit set
		return NextBit_1();
	}
}


/**********************************************************************************************/
inline bool FBL_BitSetIterator::FirstBit_0( void )
{
	if( mBitSet.mBytesInArray == 0 ) // No bits in the array
		return false;
	
	// Move to the beginning of the array
	FirstBit();
	
	// Return the next zero bit
	if( !((*pCurrentByte) & 0x01) )
	{
		// First bit is zero
		return true;
	}
	else
	{
		// Return the next zero bit
		return NextBit_0();
	}
}


#pragma mark -
 

/**********************************************************************************************/
inline bool FBL_BitSetIterator::LastBit_1( void )
{
	if( mBitSet.mBytesInArray == 0 ) // No bits in the array
		return false;
	
	// Move to the end of the array
	LastBit();
	
	// Return the next set bit
	if( (*pCurrentByte) & (0x01 << currBit) )
	{
		// Last bit is set
		return true;
	}
	else
	{
		// Return the next bit set
		return PrevBit_1();
	}
}


/**********************************************************************************************/
inline bool FBL_BitSetIterator::LastBit_0( void )
{
	if( mBitSet.mBytesInArray == 0 ) // No bits in the array
		return false;
	
	// Move to the end of the array
	LastBit();
	
	// Return the next set bit
	if( !((*pCurrentByte) & (0x01 << currBit)) )
	{
		// Last bit is set
		return true;
	}
	else
	{
		// Return the next bit set
		return PrevBit_0();
	}
}


#pragma mark -
 

/***********************************************************************************************
FUNCTION   : bool FBL_BitSetIterator::NextBit_1( void );

CALLED BY  : Next();

DESCRIPTION: Move iterator to the next nonzero bit. Returns true if such bit found.
***********************************************************************************************/
inline bool FBL_BitSetIterator::NextBit_1( void )
{
	CHECK_FBL( mBitSet.mBytesInArray > 0 ); 
	
	// We need to move to the next bit
	if( NextBit() )
	{	
		// Get the current word and adjust for the current bit
		uchar currentByte = *pCurrentByte;
		currentByte >>= currBit;
		
		// Coarse search forward until we find a byte that has at least one set bit
		while( currentByte == 0x00 )
		{
			// Work out whether we are at the real end of our array
			if( mCurrentPosition >= mBitSet.mMaxValue )
				return false;
			
			// Move to the next byte
			currentByte = *(++pCurrentByte);
			mCurrentPosition += ulong( BITS_IN_CHAR - currBit );
			currBit = 0;
		}
		
		// OK, we must have a byte here that has at least one set bit.  
		// The question is, which bit is it?
		while( currBit < BITS_IN_CHAR )
		{
			if( (currentByte & 0x01) )
			{
				break;// Found our 1 bit
			}
			else
			{
				// Work out whether we are at the real end of our array
				if( mCurrentPosition >= mBitSet.mMaxValue )
					return false;

				// Move to the next bit
				currBit++;
				mCurrentPosition++;
				currentByte >>= 1;		
			}
		}
		
		return true;
	}
	else
	{
		return false;	// No next bit
	}

} /* FBL_BitSetIterator::NextBit_1 */


/**********************************************************************************************/
inline bool FBL_BitSetIterator::NextBit_0( void )
{
	CHECK_FBL( mBitSet.mBytesInArray > 0 ); 
	
	// We need to move to the next bit
	if( NextBit() )
	{	
		// Get the current word and adjust for the current bit
		uchar currentByte = *pCurrentByte;
		currentByte >>= currBit;
		
		// Coarse search forward until we find a byte that has at least one zero bit
		while( currentByte == 0xFF )
		{
			// Work out whether we are at the real end of our array
			if( mCurrentPosition >= mBitSet.mMaxValue )
				return false;
			
			// Move to the next byte
			currentByte = *(++pCurrentByte);
			mCurrentPosition += ulong( BITS_IN_CHAR - currBit );
			currBit = 0;
		}
		
		// OK, we must have a byte here that has at least one set bit.  
		// The question is, which bit is it?
		while( currBit < BITS_IN_CHAR )
		{
			if( !(currentByte & 0x01) )
			{
				break;// Found our 0 bit
			}
			else
			{
				// Work out whether we are at the real end of our array
				if( mCurrentPosition >= mBitSet.mMaxValue )
					return false;

				// Move to the next bit
				currBit++;
				mCurrentPosition++;
				currentByte >>= 1;		
			}
		}
		
		return true;
	}
	else
	{
		return false;	// No next bit
	}
}


#pragma mark -
 

/**********************************************************************************************/
inline bool FBL_BitSetIterator::PrevBit_1( void )
{
	CHECK_FBL( mBitSet.mBytesInArray > 0 ); 
	
	// We need to move to the prev bit
	if( PrevBit() )
	{	
		// Get the current byte and adjust for the current bit
		uchar currentByte = *pCurrentByte;
		currentByte <<= uchar( BITS_IN_CHAR - (currBit + 1) );
		
		// Coarse search forward until we find a byte that has at least one set bit
		while( currentByte == 0x00 )
		{
			// Work out whether we are at the real beginning of our array
			if( mCurrentPosition <= 1 )
				return false;
			
			// Move to the prev byte
			currentByte = *(--pCurrentByte);
			mCurrentPosition -= ulong(currBit + 1);
			currBit = BITS_IN_CHAR - 1;
		}
		
		// OK, we must have a byte here that has at least one set bit.  
		// The question is, which bit is it?
		while( true )
		{
			if( (currentByte & 0x80) )
			{
				break;// Found our 1 bit
			}
			else
			{
				// Work out whether we are at the real beginning of our array
				if( mCurrentPosition <= 1 )
					return false;

				// Move to the next bit
				currBit--;
				mCurrentPosition--;
				currentByte <<= 1;		
			}
		}
		
		return true;
	}
	else
	{
		return false;	// No prev bit
	}
}


/**********************************************************************************************/
inline bool FBL_BitSetIterator::PrevBit_0( void )
{
	CHECK_FBL( mBitSet.mBytesInArray > 0 ); 
	
	// We need to move to the prev bit
	if( PrevBit() )
	{	
		// Get the current byte and adjust for the current bit
		uchar currentByte = *pCurrentByte;
		currentByte <<= uchar( BITS_IN_CHAR - (currBit + 1) );
		
		// Coarse search forward until we find a byte that has at least one set bit
		while( currentByte == 0xFF )
		{
			// Work out whether we are at the real beginning of our array
			if( mCurrentPosition <= 1 )
				return false;
			
			// Move to the prev byte
			currentByte = *(--pCurrentByte);
			mCurrentPosition -= ulong( currBit + 1 );
			currBit = BITS_IN_CHAR - 1;
		}
		
		// OK, we must have a byte here that has at least one set bit.  
		// The question is, which bit is it?
		while( true )
		{
			if( !(currentByte & 0x80) )
			{
				break;// Found our 1 bit
			}
			else
			{
				// Work out whether we are at the real beginning of our array
				if( mCurrentPosition <= 1 )
					return false;

				// Move to the next bit
				currBit--;
				mCurrentPosition--;
				currentByte <<= 1;		
			}
		}
		
		return true;
	}
	else
	{
		return false;	// No prev bit
	}
}


#pragma mark -

 
/**********************************************************************************************/
inline ID_TYPE FBL_BitSetIterator::FirstItem( void ){
	return FBL_BitSetIterator::First() ? mCurrentPosition : 0;
}


/**********************************************************************************************/
inline ID_TYPE FBL_BitSetIterator::LastItem( void ){
	return FBL_BitSetIterator::Last() ? mCurrentPosition : 0;
}


/**********************************************************************************************/
inline ID_TYPE FBL_BitSetIterator::NextItem( void ){
	return FBL_BitSetIterator::Next() ? mCurrentPosition : 0;
}


/**********************************************************************************************/
inline ID_TYPE FBL_BitSetIterator::PrevItem( void ){
	return FBL_BitSetIterator::Prev() ? mCurrentPosition : 0;
}


/**********************************************************************************************/
#if defined(__MWERKS__)
	#pragma options align=reset
#endif

#endif
